import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="157",
    message="Child is aged 4 years or over at the beginning of the year or 16 years or under at the end of the "
    "year and Strengths and Difficulties Questionnaire (SDQ) 1 has been recorded as the reason for no "
    "Strengths and Difficulties Questionnaire (SDQ) score.",
    affected_fields=["SDQ_REASON", "DOB"],
    tables=["OC2", "Episodes"],
)
def validate(dfs):
    if "OC2" not in dfs or "Episodes" not in dfs:
        return {}
    oc2 = add_CLA_column(dfs, "OC2")

    start = pd.to_datetime(
        dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
    )
    endo = pd.to_datetime(
        dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
    )
    oc2["DOB"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")

    ERRRR = (
        oc2["CONTINUOUSLY_LOOKED_AFTER"]
        & (oc2["DOB"] + pd.DateOffset(years=4) <= start)
        & (oc2["DOB"] + pd.DateOffset(years=16) >= endo)
        & oc2["SDQ_SCORE"].isna()
        & (oc2["SDQ_REASON"] == "SDQ1")
    )

    return {"OC2": oc2[ERRRR].index.to_list()}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "2",
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "3333",
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "4",
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },
            {
                "CHILD": "6666",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": "6666",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },
            {
                "CHILD": "7777",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "7777",
                "DECOM": "01/01/1981",
                "DEC": "01/07/1981",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },
            {
                "CHILD": "8888",
                "DECOM": "01/01/1981",
                "DEC": "31/03/1999",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },
        ]
    )

    oc2 = pd.DataFrame(
        {
            "CHILD": ["9999999999", "1", "2", "3333", "99999999", "8888", "5"],
            "SDQ_SCORE": ["OK", pd.NA, pd.NA, pd.NA, pd.NA, "OK", pd.NA],
            "SDQ_REASON": ["SDQ1", "SDQ1", "SDQ1", "OK", pd.NA, "OK", "OK!"],
            "DOB": [
                "01/04/1970",
                "31/03/1965",
                "01/04/1976",
                "01/04/1970",
                "01/04/1970",
                "01/04/1970",
                "01/07/1970",
            ],
        }
    )

    test_dfs = {"Episodes": eps, "OC2": oc2, "metadata": metadata}

    test_result = validate(test_dfs)

    assert test_result == {"OC2": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1014",
    message="UASC information is not required for care leavers",
    affected_fields=["ACTIV", "ACCOM", "IN_TOUCH", "DUC"],
    tables=["Episodes", "UASC", "OC3"],
)
def validate(dfs):
    if "UASC" not in dfs or "Episodes" not in dfs or "OC3" not in dfs:
        return {}
    else:
        uasc = dfs["UASC"]
        episodes = dfs["Episodes"]
        oc3 = dfs["OC3"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_end = dfs["metadata"]["collection_end"]

        # prepare to merge
        oc3.reset_index(inplace=True)
        uasc.reset_index(inplace=True)
        episodes.reset_index(inplace=True)

        collection_start = pd.to_datetime(
            collection_start, format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )

        date_check = (
            (episodes["DEC"] >= collection_start)
            & (episodes["DECOM"] <= collection_end)
        ) | ((episodes["DECOM"] <= collection_end) & episodes["DEC"].isna())
        episodes["EPS"] = date_check
        episodes["EPS_COUNT"] = episodes.groupby("CHILD")["EPS"].transform("sum")

        # inner merge to take only episodes of children which are also found on the uasc table
        merged = episodes.merge(
            uasc, on="CHILD", how="inner", suffixes=["_eps", "_sc"]
        ).merge(oc3, on="CHILD", how="left")
        # adding suffixes with the secondary merge here does not go so well yet.

        some_provided = (
            merged["ACTIV"].notna()
            | merged["ACCOM"].notna()
            | merged["IN_TOUCH"].notna()
        )

        mask = (merged["EPS_COUNT"] == 0) & some_provided

        error_locs_uasc = merged.loc[mask, "index_sc"]
        error_locs_oc3 = merged.loc[mask, "index"]

        return {
            "UASC": error_locs_uasc.unique().tolist(),
            "OC3": error_locs_oc3.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_uasc = pd.DataFrame(
        [
            {"CHILD": 101, "DOB": "01/06/2000", "DUC": "05/06/2019"},  # 0
            {"CHILD": 102, "DOB": "02/06/2000", "DUC": pd.NA},  # 1
            {"CHILD": 105, "DOB": "03/06/2000", "DUC": "01/06/2015"},  # 2
            {"CHILD": 107, "DOB": "04/06/2000", "DUC": "02/06/2020"},  # 3
            {"CHILD": 110, "DOB": pd.NA, "DUC": "05/06/2020"},  # 4 Fails
            {"CHILD": 111, "DOB": pd.NA, "DUC": "05/06/2020"},  # 5 Fails
        ]
    )
    fake_data_oc3 = pd.DataFrame(
        {
            "CHILD": [101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111],
            "IN_TOUCH": [
                pd.NA,
                "YES",
                "YES",
                pd.NA,
                "Yes",
                "No",
                "YES",
                "YES",
                pd.NA,
                pd.NA,
                "!!",
            ],
            "ACTIV": [
                pd.NA,
                pd.NA,
                "XXX",
                pd.NA,
                "XXX",
                pd.NA,
                pd.NA,
                "XXX",
                pd.NA,
                "XXX",
                "!!",
            ],
            "ACCOM": [
                pd.NA,
                pd.NA,
                pd.NA,
                "XXX",
                "XXX",
                pd.NA,
                pd.NA,
                pd.NA,
                "XXX",
                pd.NA,
                pd.NA,
            ],
        }
    )
    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/01/2020",
                "DEC": "01/01/2020",
            },  # 0
            {
                "CHILD": 102,
                "DECOM": "11/01/2020",
                "DEC": "11/01/2020",
            },  # 1
            {
                "CHILD": 103,
                "DECOM": "30/03/2020",
                "DEC": "30/03/2020",
            },  # 2
            {
                "CHILD": 104,
                "DECOM": "01/01/2020",
                "DEC": "01/01/2020",
            },  # 3
            {
                "CHILD": 105,
                "DECOM": "11/05/2020",
                "DEC": "11/05/2020",
            },  # 4 eps in range
            {
                "CHILD": 105,
                "DECOM": "01/01/2020",
                "DEC": "01/01/2020",
            },  # 5
            {
                "CHILD": 106,
                "DECOM": "22/01/2020",
                "DEC": "22/01/2020",
            },  # 6
            {
                "CHILD": 107,
                "DECOM": "11/01/2020",
                "DEC": "11/01/2020",
            },  # 7
            {
                "CHILD": 108,
                "DECOM": "22/01/2020",
                "DEC": "22/01/2020",
            },  # 8
            {
                "CHILD": 109,
                "DECOM": "25/03/2020",
                "DEC": "25/03/2020",
            },  # 9
            {
                "CHILD": 110,
                "DECOM": "01/01/2020",
                "DEC": "01/01/2020",
            },  # 10 fail.
            {
                "CHILD": 110,
                "DECOM": "01/11/2021",
                "DEC": "01/11/2021",
            },  # 11
            {
                "CHILD": 111,
                "DECOM": "01/11/2019",
                "DEC": "31/03/2021",
            },  # 12
        ]
    )

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}
    fake_dfs = {
        "UASC": fake_data_uasc,
        "Episodes": fake_data_episodes,
        "OC3": fake_data_oc3,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    # assert result == {'UASC': [2], 'Episodes': [4,5], 'OC3':[4]}
    assert result == {"UASC": [1, 3, 4], "OC3": [1, 6, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="202",
    message="The gender code conflicts with the gender already recorded for this child.",
    affected_fields=["SEX"],
    tables=["Header", "Header_last"],
)
def validate(dfs):
    if "Header" not in dfs or "Header_last" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        header_last = dfs["Header_last"]

        header_merged = (
            header.reset_index()
            .merge(
                header_last,
                how="left",
                on=["CHILD"],
                suffixes=("", "_last"),
                indicator=True,
            )
            .set_index("index")
        )

        in_both_years = header_merged["_merge"] == "both"
        sex_is_different = header_merged["SEX"].astype(str) != header_merged[
            "SEX_last"
        ].astype(str)

        error_mask = in_both_years & sex_is_different

        error_locations = header_merged.index[error_mask]

        return {"Header": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "109", "110"],
            "SEX": ["1", 2, "1", "2", pd.NA, "1", pd.NA, "2", "3"],
        }
    )

    fake_data_prev = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "107", "108", "109", "110"],
            "SEX": ["1", 1, "2", 2, pd.NA, "1", "2", pd.NA, "2"],
        }
    )

    fake_dfs = {"Header": fake_data, "Header_last": fake_data_prev}

    result = validate(fake_dfs)

    assert result == {"Header": [1, 2, 6, 7, 8]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="378",
    message="A child who is placed with parent(s) cannot be looked after under a single period of accommodation under Section 20 of the Children Act 1989.",
    affected_fields=["PLACE", "LS"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        # the & sign supercedes the ==, so brackets are necessary here
        mask = (episodes["PLACE"] == "P1") & (episodes["LS"] == "V2")
        error_locations = episodes.index[mask]
        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "11", "PLACE": "P1", "LS": "x", "REC": "x"},  # 0
            {"CHILD": "202", "PLACE": "x", "LS": "D1", "REC": "x"},  # 1
            {"CHILD": "3003", "PLACE": "A4", "LS": "D1", "REC": "E12"},  # 2
            {"CHILD": "40004", "PLACE": "P1", "LS": "V2", "REC": "E12"},  # 3
            {"CHILD": "5005", "PLACE": "A5", "LS": "x", "REC": "x"},  # 4
            {"CHILD": "606", "PLACE": "A6", "LS": "V2", "REC": "x"},  # 5
            {"CHILD": "77", "PLACE": "x", "LS": "x", "REC": "x"},  # 6
        ]
    )
    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)
    assert result == {"Episodes": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="435",
    message="Reason for new episode is that child’s placement has changed but not the legal status, "
    + "but this is not reflected in the episode data recorded.",
    affected_fields=["LS", "PLACE", "PL_POST", "URN", "PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi.sort_values(["CHILD", "DECOM"], inplace=True)
        epi["idx_orig"] = epi.index
        epi.reset_index(inplace=True)
        epi["idx_ordered"] = epi.index
        epi["idx_previous"] = epi.index + 1

        err_co = (
            epi.merge(
                epi,
                how="inner",
                left_on="idx_ordered",
                right_on="idx_previous",
                suffixes=["", "_PRE"],
            )
            .query("RNE.isin(['P', 'T']) & CHILD == CHILD_PRE")
            .query(
                "(LS != LS_PRE) | ((PLACE == PLACE_PRE) & (PL_POST == PL_POST_PRE) & (URN == URN_PRE) & "
                + "(PLACE_PROVIDER == PLACE_PROVIDER_PRE))"
            )
        )

        err_list = err_co["idx_orig"].unique().tolist()
        err_list.sort()

        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX1",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 0
            {
                "CHILD": "111",
                "DECOM": "03/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX3",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "04/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX3",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 2 Fail all the same
            {
                "CHILD": "111",
                "DECOM": "05/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX1",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 3
            {
                "CHILD": "222",
                "DECOM": "07/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX1",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR4",
            },  # 4
            {
                "CHILD": "222",
                "DECOM": "08/06/2020",
                "RNE": "P",
                "LS": "L1",
                "PL_POST": "XX1",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 5
            {
                "CHILD": "222",
                "DECOM": "09/06/2020",
                "RNE": "P",
                "LS": "L3",
                "PL_POST": "XX1",
                "URN": "SC112",
                "PLACE": "R1",
                "PLACE_PROVIDER": "PR1",
            },  # 6 Fail dif LS
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 6]}



import pandas as pd
import numpy as np

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="347",
    message="Care leavers aged under 18 years, can not be accommodated in a Foyer.",
    affected_fields=["PLACE", "DEC", "URN"],
    tables=["OC3"],
)
def validate(dfs):
    # If <ACCOM> = ‘T’ then <DOB> must be >= 18 years prior to <COLLECTION_START_DATE>
    if "OC3" not in dfs:
        return {}
    else:
        df = dfs["OC3"]
        collection_start = dfs["metadata"]["collection_start"]
        collection_start = pd.to_datetime(collection_start, format="%d/%m/%Y")

        df = df.reset_index()

        df["DOB"] = pd.to_datetime(df["DOB"], format="%d/%m/%Y")
        df["AGE_AT_CS"] = collection_start - df["DOB"]
        df_T = df[df["ACCOM"] == "T"]

        df_errors = df_T[(df_T["AGE_AT_CS"] < np.timedelta64(18, "Y"))]

        return {"OC3": df_errors["index"].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {
                "CHILD": "child1",
                "DOB": "01/01/1999",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "T",
            },  # 0 Pass over 18
            {
                "CHILD": "child2",
                "DOB": "01/01/2010",
                "IN_TOUCH": pd.NA,
                "ACTIV": pd.NA,
                "ACCOM": pd.NA,
            },  # 1 Pass under 18 but not accom = T
            {
                "CHILD": "child3",
                "DOB": "01/01/2000",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "T",
            },  # 2 Pass over 18
            {
                "CHILD": "child4",
                "DOB": "02/04/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "T",
            },  # Fail, under 18 before collection start, accom = T
            {
                "CHILD": "child5",
                "DOB": "01/04/2002",
                "IN_TOUCH": "Y",
                "ACTIV": "Y",
                "ACCOM": "T",
            },  # Pass, 18 day of collection start
        ]
    )

    fake_data["DOB"] = pd.to_datetime(fake_data["DOB"], format="%d/%m/%Y")
    metadata = {"collection_start": pd.to_datetime("01/04/2020", format="%d/%m/%Y")}

    fake_dfs = {
        "OC3": fake_data,
        "metadata": metadata,
    }

    result = validate(fake_dfs)

    assert result == {"OC3": [3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="229",
    message="Placement provider does not match between the placing authority and the local authority code of "
    "the provider. [NOTE: The provider's LA code is inferred from the its postcode, and may "
    "be inaccurate in some cases.]",
    affected_fields=["URN", "PLACE_PROVIDER"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]
        local_authority = dfs["metadata"]["localAuthority"]

        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[episodes["URN"].notna() & (episodes["URN"] != "XXXXXXX")]
        merged = episodes.merge(provider_info, on="URN", how="left")

        # If Ofsted URN is provided and not 'XXXXXXX' then: If <PLACE_PROVIDER> = 'PR1' then <LA> must equal Ofsted URN lookup <LA code>. If <PLACE_PROVIDER> = 'PR2' then Ofsted URN lookup <LA code> must not equal <LA>.
        mask = (
            merged["PLACE_PROVIDER"].eq("PR1")
            & merged["LA_CODE_INFERRED"].ne(local_authority)
        ) | (
            merged["PLACE_PROVIDER"].eq("PR2")
            & merged["LA_CODE_INFERRED"].eq(local_authority)
        )

        eps_error_locations = merged.loc[mask, "index_eps"].tolist()
        return {"Episodes": eps_error_locations}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR1",
                "URN": 1,
            },  # 0 fail
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR2",
                "URN": pd.NA,
            },  # 1 ignore: URN is nan
            {
                "CHILD": "1111",
                "PLACE_PROVIDER": "PR2",
                "URN": 4,
            },  # 2 pass
            {
                "CHILD": "2222",
                "PLACE_PROVIDER": "PR1",
                "URN": "XXXXXXX",
            },  # 3 ignore: URN
            {
                "CHILD": "3333",
                "PLACE_PROVIDER": "PR2",
                "URN": 2,
            },  # 4  fail
            {
                "CHILD": "3333",
                "PLACE_PROVIDER": "PR1",
                "URN": 2,
            },  # 5  pass
            {
                "CHILD": "4444",
                "PLACE_PROVIDER": "PR2",
                "URN": 3,
            },  # 6 pass
            {
                "CHILD": "5555",
                "PLACE_PROVIDER": "PR1",
                "URN": 4,
            },  # 7 fail
            {
                "CHILD": "5555",
                "PLACE_PROVIDER": "PR2",
                "URN": 1,
            },  # 8 pass
        ]
    )
    fake_provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "LA_CODE_INFERRED": "other",
            },  # 0
            {
                "URN": 2,
                "LA_CODE_INFERRED": "auth",
            },  # 1
            {
                "URN": 3,
                "LA_CODE_INFERRED": "other",
            },  # 2
            {
                "URN": 4,
                "LA_CODE_INFERRED": pd.NA,
            },  # 3
        ]
    )
    metadata = {"localAuthority": "auth", "provider_info": fake_provider_info}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 4, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="387",
    message="Reason episode ceased is child moved into independent living arrangement, but the child is aged under 14.",
    affected_fields=["REC"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        episodes["DEC"] = pd.to_datetime(
            episodes["DEC"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB14"] = header["DOB"] + pd.DateOffset(years=14)

        episodes_merged = (
            episodes.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        ceased_indep = episodes_merged["REC"].str.upper().astype(str).isin(["E5", "E6"])
        ceased_over_14 = episodes_merged["DOB14"] <= episodes_merged["DEC"]
        dec_present = episodes_merged["DEC"].notna()

        error_mask = ceased_indep & ~ceased_over_14 & dec_present

        error_locations = episodes_merged.index[error_mask].unique()

        return {"Episodes": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "104", "105"],
            "REC": ["E5", "E6", "X1", pd.NA, "E6", "E5", "E6"],
            "DEC": [
                "16/03/2021",
                "17/06/2020",
                "20/03/2020",
                pd.NA,
                "23/08/2020",
                "19XX/33rd",
                pd.NA,
            ],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "105"],
            "DOB": [
                "16/03/2005",
                "23/09/2002",
                "31/12/2014",
                "31/12/2014",
                "01/01/2004",
                "01/01/2004",
            ],
        }
    )

    fake_metadata = {"collection_end": "31/03/2021"}

    fake_dfs = {
        "Episodes": fake_data,
        "Header": fake_data_child,
        "metadata": fake_metadata,
    }

    result = validate(fake_dfs)

    assert result == {"Episodes": [4]}


# -------------------------------
# Tests for 452, 453, 503G & 503H all use these dataframes:

fake_eps__452_453_503G_503H_prev = pd.DataFrame(
    [
        {
            "CHILD": "101",
            "RNE": "S",
            "DECOM": "01/04/2020",
            "DEC": "04/06/2020",
            "PL_DISTANCE": 400,
            "PL_LA": "WAL",
        },
        {
            "CHILD": "101",
            "RNE": "P",
            "DECOM": "04/06/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": 400,
            "PL_LA": "WAL",
        },
        {
            "CHILD": "103",
            "RNE": "P",
            "DECOM": "03/03/2020",
            "DEC": "10/07/2020",
            "PL_DISTANCE": 10,
            "PL_LA": "816",
        },
        {
            "CHILD": "103",
            "RNE": "L",
            "DECOM": "10/07/2020",
            "DEC": "09/12/2020",
            "PL_DISTANCE": 10,
            "PL_LA": "816",
        },
        {
            "CHILD": "104",
            "RNE": "B",
            "DECOM": "28/03/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": 185,
            "PL_LA": "356",
        },
        {
            "CHILD": "105",
            "RNE": "L",
            "DECOM": "16/04/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": "165",
            "PL_LA": "112",
        },
        {
            "CHILD": "106",
            "RNE": "S",
            "DECOM": "04/11/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": "165",
            "PL_LA": "112",
        },
    ]
)

fake_eps__452_453_503G_503H = pd.DataFrame(
    [
        {
            "CHILD": "101",
            "RNE": "P",
            "DECOM": "04/06/2020",
            "DEC": "04/06/2021",
            "PL_DISTANCE": 400,
            "PL_LA": "WAL",
        },
        {
            "CHILD": "101",
            "RNE": "P",
            "DECOM": "04/06/2021",
            "DEC": pd.NA,
            "PL_DISTANCE": 400,
            "PL_LA": "WAL",
        },
        {
            "CHILD": "102",
            "RNE": "L",
            "DECOM": "20/12/2020",
            "DEC": "07/04/2021",
            "PL_DISTANCE": 10,
            "PL_LA": "816",
        },
        # Ignore all
        {
            "CHILD": "103",
            "RNE": "L",
            "DECOM": "02/02/2021",
            "DEC": pd.NA,
            "PL_DISTANCE": 10,
            "PL_LA": "816",
        },  # Ignore all
        {
            "CHILD": "104",
            "RNE": "B",
            "DECOM": "28/03/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": 999.9,
            "PL_LA": "CON",
        },  # Fail all
        {
            "CHILD": "105",
            "RNE": "L",
            "DECOM": "16/04/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": 165,
            "PL_LA": 112,
        },
        {
            "CHILD": "106",
            "RNE": "S",
            "DECOM": "04/11/2020",
            "DEC": pd.NA,
            "PL_DISTANCE": pd.NA,
            "PL_LA": pd.NA,
        },
        # Fail 452 and 503G only
    ]
)

fake_dfs__452_453_503G_503H = {
    "Episodes": fake_eps__452_453_503G_503H,
    "Episodes_last": fake_eps__452_453_503G_503H_prev,
}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="3001",
    message="Where care leavers information is being returned for a young person around their 17th birthday, the accommodation cannot be with their former foster carer(s).",
    affected_fields=["REC"],
    tables=["Header", "OC3"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    if "OC3" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        oc3 = dfs["OC3"]
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )

        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB17"] = header["DOB"] + pd.DateOffset(years=17)

        oc3_merged = (
            oc3.reset_index()
            .merge(
                header,
                how="left",
                on=["CHILD"],
                suffixes=("", "_header"),
                indicator=True,
            )
            .set_index("index")
        )

        accom_foster = oc3_merged["ACCOM"].str.upper().astype(str).isin(["Z1", "Z2"])
        age_17_in_year = (oc3_merged["DOB17"] <= collection_end) & (
            oc3_merged["DOB17"] >= collection_start
        )

        error_mask = accom_foster & age_17_in_year

        error_locations = oc3.index[error_mask]

        return {"OC3": error_locations.to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "CHILD": ["101", "102", "101", "102", "103", "102"],
            "ACCOM": ["Z1", "Z2", "T1", pd.NA, "Z1", "Z3"],
        }
    )

    fake_data_child = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103"],
            "DOB": ["16/03/2004", "23/09/2003", "31/12/2006"],
        }
    )

    metadata = {"collection_start": "01/04/2020", "collection_end": "31/03/2021"}

    fake_dfs = {"OC3": fake_data, "Header": fake_data_child, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"OC3": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="EPI",
    message="WARNING: Episodes need to be loaded for this child before further validation is possible "
    "[NOTE: This refers to the DfE portal - here, all checks that can be performed with only the "
    "available data will be.]",
    affected_fields=["CHILD"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]

        if "Episodes" not in dfs:
            episodes = pd.DataFrame(columns=["CHILD"])
        else:
            episodes = dfs["Episodes"]

        if "OC3" not in dfs:
            oc3 = pd.DataFrame(columns=["CHILD"])
        else:
            oc3 = dfs["OC3"]

        header["index_header"] = header.index

        merged = header.merge(
            episodes["CHILD"],
            on="CHILD",
            indicator="_merge_eps",
            how="left",
            suffixes=["", "_episodes"],
        )

        merged = merged.merge(
            oc3["CHILD"],
            on="CHILD",
            indicator="_merge_oc3",
            how="left",
            suffixes=["", "_oc3"],
        )

        mask = (merged["_merge_eps"] == "left_only") & (
            merged["_merge_oc3"] == "left_only"
        )
        eps_error_locations = merged.loc[mask, "index_header"]
        return {"Header": eps_error_locations.unique().tolist()}


def test_validate():
    import pandas as pd

    fake_header = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "106"],
        }
    )

    fake_episodes = pd.DataFrame(
        {
            "CHILD": ["101", "101", "103"],
        }
    )

    fake_dfs = {"Header": fake_header, "Episodes": fake_episodes}
    result = validate(fake_dfs)
    assert result == {"Header": [1, 3]}


# Testing rule with Header, Episodes and OC3 provided.



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="377",
    message="Only two temporary placements coded as being due to holiday of usual foster carer(s) are "
    + "allowed in any 12- month period.",
    affected_fields=["PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi.reset_index(inplace=True)

        potent_cohort = epi[epi["PLACE"] == "T3"]

        # Here I'm after the CHILD ids where there are more than 2 T3 placements.
        count_them = potent_cohort.groupby("CHILD")["CHILD"].count().to_frame(name="cc")
        count_them.reset_index(inplace=True)
        count_them = count_them[count_them["cc"] > 2]

        err_coh = epi[epi["CHILD"].isin(count_them["CHILD"])]
        err_coh = err_coh[err_coh["PLACE"] == "T3"]

        err_list = err_coh["index"].unique().tolist()
        err_list.sort()
        return {"Episodes": err_list}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "111", "RNE": "P", "DECOM": "01/06/2020", "PLACE": "T3"},  # 0 ^
            {"CHILD": "111", "RNE": "P", "DECOM": "05/06/2020", "PLACE": "P3"},  # 1
            {"CHILD": "111", "RNE": "T", "DECOM": "06/06/2020", "PLACE": "T3"},  # 2
            {"CHILD": "111", "RNE": "B", "DECOM": "08/06/2020", "PLACE": "T3"},  # 3 v
            {"CHILD": "222", "RNE": "B", "DECOM": "05/06/2020", "PLACE": "T3"},  # 4 -
            {"CHILD": "333", "RNE": "B", "DECOM": "06/06/2020", "PLACE": "T3"},  # 5 ^
            {"CHILD": "333", "RNE": "U", "DECOM": "10/06/2020", "PLACE": "T3"},  # 6 v
            {"CHILD": "444", "RNE": "B", "DECOM": "07/06/2020", "PLACE": "T3"},  # 7 ^
            {"CHILD": "444", "RNE": "B", "DECOM": "08/06/2020", "PLACE": "oo"},  # 8
            {"CHILD": "444", "RNE": "T", "DECOM": "09/06/2020", "PLACE": "T3"},  # 9
            {"CHILD": "444", "RNE": "B", "DECOM": "15/06/2020", "PLACE": "T3"},  # 10 v
            {"CHILD": "6666", "RNE": "P", "DECOM": "11/06/2020", "PLACE": "oo"},  # 11 ^
            {"CHILD": "6666", "RNE": "P", "DECOM": "12/06/2020", "PLACE": "T3"},  # 12
            {"CHILD": "6666", "RNE": "P", "DECOM": "13/06/2020", "PLACE": "T3"},  # 13
            {"CHILD": "6666", "RNE": "P", "DECOM": "14/06/2020", "PLACE": "T3"},  # 14 v
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 2, 3, 7, 9, 10, 12, 13, 14]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="420",
    message="LA of placement completed but child is looked after under legal status V3 or V4.",
    affected_fields=["PL_LA"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        is_short_term = df["LS"].isin(["V3", "V4"])

        mask = is_short_term & df["PL_LA"].notna()
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS": ["C2", "V3", "V4", "V3", "V4", "C2"],
            "PL_LA": [pd.NA, "E03934134", "E059635", pd.NA, pd.NA, "NIR"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    assert validate(fake_dfs) == {"Episodes": [1, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="SW09STG2",
    message="Social worker episode appears to have lasted for less than 24 hours.",
    affected_fields=["SW_DECOM", "SW_DEC"],
    tables=["SWEpisodes"],
)
def validate(dfs):
    if "SWEpisodes" not in dfs:
        return {}
    else:
        df = dfs["SWEpisodes"]

        error_rows = df[df["SW_DEC"] == df["SW_DECOM"]].index

        return {"SWEpisodes": error_rows.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"CHILD": "1", "SW_DEC": "01/01/2000", "SW_DECOM": "01/01/2000"},  # fail
            {"CHILD": "1", "SW_DEC": "01/01/2000", "SW_DECOM": "02/01/2000"},
            {"CHILD": "2", "SW_DEC": "01/01/1901", "SW_DECOM": "01/01/1901"},  # fail
            {"CHILD": "2", "SW_DEC": "01/01/1901", "SW_DECOM": pd.NA},
            {"CHILD": "3", "SW_DEC": pd.NA, "SW_DECOM": pd.NA},
        ]
    )

    fake_dfs = {"SWEpisodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"SWEpisodes": [0, 2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.rules.rule_utils import dec_after_decom


@rule_definition(
    code="379",
    message="Temporary placements for unspecified reason (placement code T4) cannot exceed seven days.",
    affected_fields=["DECOM", "PLACE"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Header" not in dfs:
        return {}
    else:
        return dec_after_decom(dfs, p_code="T4", y_gap=7)


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666"],
            "DECOM": [
                "01/01/2020",
                "01/02/2020",
                "02/03/2020",
                "15/03/2010",
                "15/03/2004",
                "14/03/2014",
            ],
            "DEC": [
                "08/01/2020",
                "09/02/2020",
                "09/03/2020",
                "25/03/2010",
                "22/03/2004",
                "25/03/2014",
            ],
            "PLACE": ["T4", "T4", "T4", "T4", "T4", "P2"],
        }
    )
    fake_hea = pd.DataFrame(
        {
            "CHILD": ["111", "222", "333", "444", "555", "666"],
            "DOB": [
                "08/01/2020",
                "08/02/2020",
                "09/03/2020",
                "22/03/2010",
                "22/03/2004",
                "21/03/2014",
            ],
        }
    )

    fake_hea["DOB"] = pd.to_datetime(
        fake_hea["DOB"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DECOM"] = pd.to_datetime(
        fake_epi["DECOM"], format="%d/%m/%Y", errors="coerce"
    )
    fake_epi["DEC"] = pd.to_datetime(
        fake_epi["DEC"], format="%d/%m/%Y", errors="coerce"
    )

    fake_dfs = {"Episodes": fake_epi, "Header": fake_hea}

    assert validate(fake_dfs) == {"Episodes": [1, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="187",
    message="Child cannot be looked after continuously for 12 months at "
    + "31 March (OC2) and have any of adoption or care leavers returns completed.",
    affected_fields=[
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",  # OC3
        "IN_TOUCH",
        "ACTIV",
        "ACCOM",
    ],  # AD1
    tables=["OC3", "AD1", "Episodes"],
)
def validate(dfs):
    if "OC3" not in dfs or "AD1" not in dfs or "Episodes" not in dfs:
        return {}

    # add 'CONTINUOUSLY_LOOKED_AFTER' column
    ad1, oc3 = add_CLA_column(dfs, ["AD1", "OC3"])

    # OC3
    should_be_blank = ["IN_TOUCH", "ACTIV", "ACCOM"]
    oc3_mask = oc3["CONTINUOUSLY_LOOKED_AFTER"] & oc3[should_be_blank].notna().any(
        axis=1
    )
    oc3_error_locs = oc3[oc3_mask].index.to_list()

    # AD1
    should_be_blank = [
        "DATE_INT",
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ]
    ad1_mask = ad1["CONTINUOUSLY_LOOKED_AFTER"] & ad1[should_be_blank].notna().any(
        axis=1
    )
    ad1_error_locs = ad1[ad1_mask].index.to_list()

    return {"AD1": ad1_error_locs, "OC3": oc3_error_locs}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "2",
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "3333",
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "4",
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },
            {
                "CHILD": "6666",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },  # !! - False
            {
                "CHILD": "6666",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/01/1981",
                "DEC": "01/07/1981",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # !! - False
            {
                "CHILD": "8888",
                "DECOM": "01/01/1981",
                "DEC": "31/03/1999",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # False
            {
                "CHILD": "9",
                "DECOM": "05/04/1980",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "xx",
                "RNE": "S",
            },
            # nCLA True DECOM>col_start and RNE==S
            {
                "CHILD": "10",
                "DECOM": pd.NA,
                "DEC": "05/04/1980",
                "LS": "o",
                "REC": "xx",
                "RNE": "S",
            },
            # nCLA True DEC>col_start and REC!=X1
            {
                "CHILD": "11",
                "DECOM": pd.NA,
                "DEC": "05/04/1980",
                "LS": "V3",
                "REC": pd.NA,
                "RNE": "S",
            },
            # nCLA True DEC>col_start and LS in [V3, V4]
        ]
    )

    oc3 = pd.DataFrame(
        {
            "CHILD": [
                "9999999999",
                "1",
                "2",
                "3333",
                "99999999",
                "8888",
                "5",
                "9",
                "10",
                "11",
            ],
            "IN_TOUCH": [
                "OK",
                "!!!",
                pd.NA,
                "OK",
                pd.NA,
                pd.NA,
                pd.NA,
                "xx",
                "xx",
                "xx",
            ],
        }
    )
    other_oc3_cols = ["ACTIV", "ACCOM"]
    oc3 = oc3.assign(**{col: pd.NA for col in other_oc3_cols})

    ad1 = pd.DataFrame(
        {
            "CHILD": [
                "1",
                "2",
                "3333",
                "7777",
                "99999999",
                "8888",
                "5",
                "9",
                "10",
                "11",
            ],
            "DATE_INT": [
                pd.NA,
                pd.NA,
                "OK",
                "OK",
                pd.NA,
                pd.NA,
                "!!!",
                "xx",
                "xx",
                "xx",
            ],
        }
    )

    other_ad1_cols = [
        "DATE_MATCH",
        "FOSTER_CARE",
        "NB_ADOPTR",
        "SEX_ADOPTR",
        "LS_ADOPTR",
    ]
    ad1 = ad1.assign(**{col: pd.NA for col in other_ad1_cols})

    fake_dfs = {"Episodes": eps, "OC3": oc3, "AD1": ad1, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {
        "OC3": [
            1,
        ],
        "AD1": [
            6,
        ],
    }



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="179",
    message="Placement location code is not a valid code.",
    affected_fields=["PL_LOCATION"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        is_short_term = df["LS"].isin(["V3", "V4"])

        # Because PL_LOCATION is derived, it will always be valid if present
        mask = ~is_short_term & df["PL_LOCATION"].isna()
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS": ["C2", "C2", "C2", "V3", "V4"],
            "PL_LOCATION": [
                "IN",
                "OUT",
                pd.NA,
                pd.NA,
                pd.NA,
            ],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    assert validate(fake_dfs) == {"Episodes": [2]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="546",
    message="Children aged 5 or over at 31 March should not have health promotion information completed.",
    affected_fields=["DOB", "HEALTH_CHECK"],
    tables=["Episodes", "OC2"],
)
def validate(dfs):
    if "OC2" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        collection_end = dfs["metadata"]["collection_end"]
        # add CLA column
        oc2 = add_CLA_column(dfs, "OC2")

        # to datetime
        oc2["DOB"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )

        # If <DOB> >= 5 years prior to<COLLECTION_END_DATE>and<CONTINUOUSLY_LOOKED_AFTER>= 'Y' then<HEALTH_CHECK>` should not be provided
        mask = (
            (collection_end >= (oc2["DOB"] + pd.offsets.DateOffset(years=5)))
            & oc2["CONTINUOUSLY_LOOKED_AFTER"]
            & oc2["HEALTH_CHECK"].notna()
        )
        error_locations = oc2.index[mask]
        return {"OC2": error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": 101,
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": 102,
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": 103,
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": 104,
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": 105,
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
        ]
    )
    fake_data = pd.DataFrame(
        {
            "CHILD": [101, 102, 103, 104, 105],
            "DOB": ["08/03/1963", "22/06/1957", pd.NA, "13/10/2000", "10/01/1948"],
            "HEALTH_CHECK": [1, 1, pd.NA, 1, pd.NA],
            # 0 fail because conditions are met but HEALTH_CHECK is provided
            # 1 fail because conditions are met but HEALTH_CHECK is provided
            # 2 ignore: DOB is nan
            # 3 ignore: CLA is false
            # 4 pass
        }
    )

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    fake_dfs = {"OC2": fake_data, "metadata": metadata, "Episodes": fake_data_episodes}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 1]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="562",
    message="Episode commenced before the start of the current collection year but there is a missing continuous episode in the previous year.",
    affected_fields=["DECOM"],
    tables=["Episodes", "Episodes_last"],
)
def validate(dfs):
    if "Episodes" not in dfs or "Episodes_last" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        epi_last = dfs["Episodes_last"]
        epi["DECOM"] = pd.to_datetime(epi["DECOM"], format="%d/%m/%Y", errors="coerce")
        epi_last["DECOM"] = pd.to_datetime(
            epi_last["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        collection_start = pd.to_datetime(
            dfs["metadata"]["collection_start"], format="%d/%m/%Y", errors="coerce"
        )

        epi.reset_index(inplace=True)
        epi = epi[epi["DECOM"] < collection_start]

        grp_decom_by_child = epi.groupby(["CHILD"])["DECOM"].idxmin(skipna=True)
        min_decom = epi.loc[epi.index.isin(grp_decom_by_child), :]

        grp_last_decom_by_child = epi_last.groupby(["CHILD"])["DECOM"].idxmax(
            skipna=True
        )
        max_last_decom = epi_last.loc[epi_last.index.isin(grp_last_decom_by_child), :]

        merged_co = min_decom.merge(
            max_last_decom,
            how="left",
            on=["CHILD", "DECOM"],
            suffixes=["", "_PRE"],
            indicator=True,
        )
        error_cohort = merged_co[merged_co["_merge"] == "left_only"]

        error_list = error_cohort["index"].to_list()
        error_list = list(set(error_list))
        error_list.sort()
        return {"Episodes": error_list}


def test_validate():
    import pandas as pd

    fake_epi = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "15/03/2021"},  # 0 Min pre year start
            {"CHILD": "111", "DECOM": "05/06/2021"},  # 1
            {"CHILD": "222", "DECOM": "13/03/2021"},  # 2 Min pre year start
            {"CHILD": "222", "DECOM": "08/06/2021"},  # 3
            {"CHILD": "222", "DECOM": "05/06/2021"},  # 4
            {"CHILD": "333", "DECOM": "01/01/2021"},  # 5 Min pre year start
            {"CHILD": "444", "DECOM": "01/05/2021"},  # 6
        ]
    )
    fake_last = pd.DataFrame(
        [
            {"CHILD": "111", "DECOM": "01/06/2020"},  # 0
            {"CHILD": "111", "DECOM": "05/06/2020"},  # 1
            {"CHILD": "111", "DECOM": "15/03/2021"},  # 2 Max matches next year
            {"CHILD": "222", "DECOM": "01/02/2021"},  # 3
            {"CHILD": "222", "DECOM": "11/03/2021"},  # 4 Max doesn't match - fail
            {"CHILD": "333", "DECOM": "06/06/2020"},  # 5 Max doesn't match - fail
        ]
    )
    metadata = {"collection_start": "01/04/2021"}

    fake_dfs = {"Episodes": fake_epi, "Episodes_last": fake_last, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="408",
    message="Child is placed for adoption with a placement order, but no placement order has been recorded.",
    affected_fields=["PLACE", "LS"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        epi = dfs["Episodes"]
        error_mask = epi["PLACE"].isin(["A5", "A6"]) & (epi["LS"] != "E1")
        return {"Episodes": epi.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"PLACE": "A5", "LS": "S"},  # 0   Fail
            {"PLACE": "V4", "LS": "T"},  # 1
            {"PLACE": "A6", "LS": "E1"},  # 2
            {"PLACE": "U2", "LS": pd.NA},  # 3
            {"PLACE": "A6", "LS": "U"},  # 4  Fail
            {"PLACE": "A5", "LS": pd.NA},  # 5  Fail
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="336",
    message="Child does not have a foster placement immediately prior to being placed for adoption.",
    affected_fields=["PLACE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        episodes = dfs["Episodes"]
        place_code_list = ["A3", "A4"]
        prev_code_list = [
            "A3",
            "A4",
            "A5",
            "A6",
            "U1",
            "U2",
            "U3",
            "U4",
            "U5",
            "U6",
        ]

        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        # new column that contains place of previous episode

        sorted_and_grouped_eps = episodes.sort_values("DECOM").groupby("CHILD")

        episodes["PLACE_prev"] = sorted_and_grouped_eps["PLACE"].shift(1)
        episodes["is_first_episode"] = False
        episodes.loc[
            sorted_and_grouped_eps["DECOM"].idxmin(), "is_first_episode"
        ] = True

        # Where <PL> = 'A3' or 'A5' previous episode <PL> must be one of:
        # ('A3'; 'A4'; 'A5'; 'A6'; 'U1', 'U2', 'U3', 'U4', 'U5' or 'U6')
        mask = (
            (episodes["PLACE"].isin(place_code_list))
            & ~episodes["PLACE_prev"].isin(prev_code_list)
            & ~episodes["is_first_episode"]  # omit first eps, as prev_PLACE is NaN
        )

        # error locations
        error_locs = episodes.index[mask]
        return {"Episodes": error_locs.tolist()}


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/01/2020",
                "PLACE": "U1",
            },  # 0 ignored no previous episode
            {
                "CHILD": "111",
                "DECOM": "11/01/2020",
                "PLACE": "T1",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "22/01/2020",
                "PLACE": "A3",
            },  # 2 fail (T1 -> A3)
            {
                "CHILD": "123",
                "DECOM": "11/01/2020",
                "PLACE": "A4",
            },  # 4 pass (U1 -> A4)
            {
                "CHILD": "123",
                "DECOM": "01/01/2020",
                "PLACE": "U1",
            },  # 3
            {
                "CHILD": "333",
                "DECOM": "01/01/2020",
                "PLACE": "U1",
            },  # 5
            {
                "CHILD": "333",
                "DECOM": "22/01/2020",
                "PLACE": "A3",
            },  # 6 fail (T2 -> A3)
            {
                "CHILD": "333",
                "DECOM": "11/01/2020",
                "PLACE": "T2",
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "22/01/2020",
                "PLACE": "A4",
            },  # 8 fail (X1 -> A4)
            {
                "CHILD": "444",
                "DECOM": "11/01/2020",
                "PLACE": "X1",
            },  # 9
            {
                "CHILD": "444",
                "DECOM": "01/01/2020",
                "PLACE": "A3",
            },  # 10 ignored no previous episode
            {
                "CHILD": "666",
                "DECOM": "01/01/2020",
                "PLACE": "A5",
            },  # 11
            {
                "CHILD": "777",
                "DECOM": "11/01/2020",
                "PLACE": "A4",
            },  # 12 fail (null -> A4)
            {
                "CHILD": "777",
                "DECOM": "01/01/2020",
                "PLACE": pd.NA,
            },  # 13
        ]
    )
    fake_dfs = {"Episodes": fake_data_episodes}

    result = validate(fake_dfs)
    assert result == {"Episodes": [2, 6, 8, 12]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="366",
    message="A child cannot change placement during the course of an individual short-term respite break.",
    affected_fields=["RNE"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        mask = (df["LS"] == "V3") & (df["RNE"] != "S")
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"LS": "V3", "RNE": "S"},  # 0
            {"LS": "V4", "RNE": "T"},  # 1
            {"LS": "U1", "RNE": pd.NA},  # 2
            {"LS": "U2", "RNE": pd.NA},  # 3
            {"LS": "V3", "RNE": "U"},  # 4  Fail
            {"LS": "V3", "RNE": pd.NA},  # 5  Fail
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [4, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="542",
    message="A child aged under 10 at 31 March should not have conviction information completed.",
    affected_fields=["CONVICTED"],
    tables=["OC2s"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]
        oc2["DOB"] = pd.to_datetime(oc2["DOB"], format="%d/%m/%Y", errors="coerce")
        collection_end = pd.to_datetime(
            dfs["metadata"]["collection_end"], format="%d/%m/%Y", errors="coerce"
        )
        error_mask = (
            oc2["DOB"] + pd.offsets.DateOffset(years=10) > collection_end
        ) & oc2["CONVICTED"].notna()
        return {"OC2": oc2.index[error_mask].to_list()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DOB": ["08/03/2020", "22/06/2000", pd.NA, "13/10/2000", "10/01/2017"],
            "CONVICTED": [1, pd.NA, 1, 1, 1],  # 0 , 4
        }
    )

    metadata = {"collection_end": "31/03/2020"}

    fake_dfs = {"OC2": fake_data, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="210",
    message="Children looked after for more than a week at 31 March should not have an unknown Unique Pupil Number (UPN) code of UN4.",
    affected_fields=["UPN", "DECOM"],
    tables=["Header", "Episodes"],
)
def validate(dfs):
    if "Header" not in dfs or "Episodes" not in dfs:
        return {}
    else:
        header = dfs["Header"]
        episodes = dfs["Episodes"]
        collection_end = dfs["metadata"]["collection_end"]
        # convert to datetime
        episodes["DECOM"] = pd.to_datetime(
            episodes["DECOM"], format="%d/%m/%Y", errors="coerce"
        )
        collection_end = pd.to_datetime(
            collection_end, format="%d/%m/%Y", errors="coerce"
        )
        yr = collection_end.year
        reference_date = ref_date = pd.to_datetime(
            "24/03/" + str(yr), format="%d/%m/%Y", errors="coerce"
        )
        # prepare to merge
        episodes.reset_index(inplace=True)
        header.reset_index(inplace=True)
        # the logical way is to merge left on UPN but that will be a one to many merge and may not go as well as a many to one merge that we've been doing.
        merged = episodes.merge(
            header, on="CHILD", how="left", suffixes=["_eps", "_er"]
        )
        # If <UPN> = 'UN4' then no episode <DECOM> must be >` = 24/03/YYYY Note: YYYY refers to the current collection year.
        mask = (merged["UPN"] == "UN4") & (merged["DECOM"] >= reference_date)
        # error locations
        error_locs_header = merged.loc[mask, "index_er"]
        error_locs_eps = merged.loc[mask, "index_eps"]
        return {
            "Episodes": error_locs_eps.tolist(),
            "Header": error_locs_header.unique().tolist(),
        }


def test_validate():
    import pandas as pd

    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["111", "123", "333", "444", "667"],
            "UPN": [
                "UN4",
                pd.NA,
                "UN3",
                "UN4",
                "UN4",
            ],
        }
    )
    fake_data_epi = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "DECOM": "01/01/2020",
            },  # 0
            {
                "CHILD": "111",
                "DECOM": "11/01/2020",
            },  # 1
            {
                "CHILD": "111",
                "DECOM": "30/03/2020",
            },  # 2 fail
            {
                "CHILD": "123",
                "DECOM": "01/01/2020",
            },  # 3
            {
                "CHILD": "123",
                "DECOM": "11/01/2020",
            },  # 4
            {
                "CHILD": "333",
                "DECOM": "01/01/2020",
            },  # 5
            {
                "CHILD": "333",
                "DECOM": "22/01/2020",
            },  # 6
            {
                "CHILD": "333",
                "DECOM": "11/01/2020",
            },  # 7
            {
                "CHILD": "444",
                "DECOM": "22/01/2020",
            },  # 8
            {
                "CHILD": "444",
                "DECOM": "25/03/2020",
            },  # 9 fail
            {
                "CHILD": "444",
                "DECOM": "01/01/2020",
            },  # 10
            {
                "CHILD": "667",
                "DECOM": "01/01/2020",
            },  # 11
        ]
    )
    metadata = {"collection_end": "31/03/2020"}

    fake_dfs = {
        "Episodes": fake_data_epi,
        "Header": fake_data_header,
        "metadata": metadata,
    }

    result = validate(fake_dfs)
    assert result == {"Episodes": [2, 9], "Header": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="103",
    message="The ethnicity code is either not valid or has not been entered.",
    affected_fields=["ETHNIC"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}

    header = dfs["Header"]
    code_list = [
        "WBRI",
        "WIRI",
        "WOTH",
        "WIRT",
        "WROM",
        "MWBC",
        "MWBA",
        "MWAS",
        "MOTH",
        "AIND",
        "APKN",
        "ABAN",
        "AOTH",
        "BCRB",
        "BAFR",
        "BOTH",
        "CHNE",
        "OOTH",
        "REFU",
        "NOBT",
    ]

    mask = header["ETHNIC"].isin(code_list)

    validation_error_mask = ~mask
    validation_error_locations = header.index[validation_error_mask]

    return {"Header": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "ETHNIC": ["WBRI", "WIRI", "WOTH", "wbri", "White British", "", pd.NA],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [3, 4, 5, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition
from lac_validator.utils import (
    add_col_to_tables_CONTINUOUSLY_LOOKED_AFTER as add_CLA_column,  # Check 'Episodes' present before use!
)


@rule_definition(
    code="190",
    message="Child has not been looked after continuously for at least 12 months at 31 March but one or more "
    "data items relating to children looked after for 12 months have been completed.",
    affected_fields=[
        "CONVICTED",
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ],  # AD1
    tables=["OC2", "Episodes"],
)
def validate(dfs):
    if "OC2" not in dfs or "Episodes" not in dfs:
        return {}

    # add 'CONTINUOUSLY_LOOKED_AFTER' column
    oc2 = add_CLA_column(dfs, "OC2")

    should_be_blank = [
        "CONVICTED",
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ]

    mask = ~oc2["CONTINUOUSLY_LOOKED_AFTER"] & oc2[should_be_blank].notna().any(axis=1)
    error_locs = oc2[mask].index.to_list()

    return {"OC2": error_locs}


def test_validate():
    import pandas as pd

    metadata = {"collection_start": "01/04/1980", "collection_end": "31/03/1981"}

    eps = pd.DataFrame(
        [
            {
                "CHILD": "1",
                "DECOM": "01/03/1980",
                "DEC": "31/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "2",
                "DECOM": "01/03/1980",
                "DEC": "30/03/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "3333",
                "DECOM": "01/03/1980",
                "DEC": "01/01/1981",
                "LS": "V3",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "4",
                "DECOM": "01/02/1970",
                "DEC": pd.NA,
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },
            {
                "CHILD": "5",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },
            {
                "CHILD": "6666",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "!!",
                "RNE": "o",
            },  # !! - False
            {
                "CHILD": "6666",
                "DECOM": "01/01/1981",
                "DEC": pd.NA,
                "LS": "o",
                "REC": pd.NA,
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/03/1979",
                "DEC": "01/01/1981",
                "LS": "o",
                "REC": "X1",
                "RNE": "o",
            },  # False
            {
                "CHILD": "7777",
                "DECOM": "01/01/1981",
                "DEC": "01/07/1981",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # !! - False
            {
                "CHILD": "8888",
                "DECOM": "01/01/1981",
                "DEC": "31/03/1999",
                "LS": "o",
                "REC": "o",
                "RNE": "S",
            },  # False
        ]
    )

    oc2 = pd.DataFrame(
        {
            "CHILD": ["9999999999", "1", "2", "3333", "99999999", "8888", "5"],
            "CONVICTED": ["!!!", "OK", pd.NA, pd.NA, pd.NA, "!!!", "OK"],
        }
    )
    other_oc2_cols = [
        "HEALTH_CHECK",
        "IMMUNISATIONS",
        "TEETH_CHECK",
        "HEALTH_ASSESSMENT",
        "SUBSTANCE_MISUSE",
        "INTERVENTION_RECEIVED",
        "INTERVENTION_OFFERED",
    ]
    oc2 = oc2.assign(**{col: pd.NA for col in other_oc2_cols})

    fake_dfs = {"Episodes": eps, "OC2": oc2, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 5]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="214",
    message="Placement location information not required.",
    affected_fields=["PL_POST", "URN"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}
    else:
        df = dfs["Episodes"]
        mask = df["LS"].isin(["V3", "V4"]) & (
            (df["PL_POST"].notna()) | (df["URN"].notna())
        )
        return {"Episodes": df.index[mask].tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        [
            {"LS": "V3", "PL_POST": "M2 3RT", "URN": "XXXXXX"},  # 0  Fail
            {"LS": "U1", "PL_POST": "M2 3RT", "URN": "SC045099"},  # 1
            {"LS": "U3", "PL_POST": pd.NA, "URN": "SC045099"},  # 2
            {"LS": "V4", "PL_POST": "M2 3RT", "URN": pd.NA},  # 3  Fail
            {"LS": "V4", "PL_POST": pd.NA, "URN": "SC045099"},  # 4  Fail
            {"LS": "T1", "PL_POST": "M2 3RT", "URN": "SC045100"},  # 5
            {"LS": "U6", "PL_POST": "M2 3RT", "URN": "SC045101"},  # 6
            {"LS": "V3", "PL_POST": pd.NA, "URN": pd.NA},  # 7
        ]
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [0, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="177",
    message="The legal status of adopter(s) code is not a valid code.",
    affected_fields=["LS_ADOPTR"],
    tables=["AD1"],
)
def validate(dfs):
    if "AD1" not in dfs:
        return {}

    adoptions = dfs["AD1"]
    code_list = ["L0", "L11", "L12", "L2", "L3", "L4"]

    mask = adoptions["LS_ADOPTR"].isin(code_list) | adoptions["LS_ADOPTR"].isna()

    validation_error_mask = ~mask
    validation_error_locations = adoptions.index[validation_error_mask]

    return {"AD1": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "LS_ADOPTR": ["L0", "L11", "L1", "l2", "", pd.NA],
        }
    )

    fake_dfs = {"AD1": fake_data}

    result = validate(fake_dfs)

    assert result == {"AD1": [2, 3, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="1006",
    message="Missing type invalid.",
    affected_fields=["MISSING"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}

    missing_from_care = dfs["Missing"]
    code_list = ["M", "A"]

    mask = (
        missing_from_care["MISSING"].isin(code_list)
        | missing_from_care["MISSING"].isna()
    )

    validation_error_mask = ~mask
    validation_error_locations = missing_from_care.index[validation_error_mask]

    return {"Missing": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "MISSING": ["M", "A", "AWAY", "NA", "", pd.NA, "m"],
        }
    )

    fake_dfs = {"Missing": fake_data}

    result = validate(fake_dfs)

    assert result == {"Missing": [2, 3, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="574",
    message="A new missing/away from placement without authorisation period cannot start when the previous missing/away from placement without authorisation period is still open. Missing/away from placement without authorisation periods should also not overlap.",
    affected_fields=["MIS_START", "MIS_END"],
    tables=["Missing"],
)
def validate(dfs):
    if "Missing" not in dfs:
        return {}
    else:
        mis = dfs["Missing"]
        mis["MIS_START"] = pd.to_datetime(
            mis["MIS_START"], format="%d/%m/%Y", errors="coerce"
        )
        mis["MIS_END"] = pd.to_datetime(
            mis["MIS_END"], format="%d/%m/%Y", errors="coerce"
        )

        mis["MIS_END_FILL"] = mis["MIS_END"].fillna(mis["MIS_START"])
        mis.sort_values(["CHILD", "MIS_END_FILL", "MIS_START"], inplace=True)

        mis.reset_index(inplace=True)
        mis.reset_index(inplace=True)  # Twice on purpose

        mis["LAG_INDEX"] = mis["level_0"].shift(-1)

        lag_mis = mis.merge(
            mis,
            how="inner",
            left_on="level_0",
            right_on="LAG_INDEX",
            suffixes=["", "_PREV"],
        )

        # We're only interested in cases where there is more than one row for a child.
        lag_mis = lag_mis[lag_mis["CHILD"] == lag_mis["CHILD_PREV"]]

        # A previous MIS_END date is null
        mask1 = lag_mis["MIS_END_PREV"].isna()
        # MIS_START is before previous MIS_END (overlapping dates)
        mask2 = lag_mis["MIS_START"] < lag_mis["MIS_END_PREV"]

        mask = mask1 | mask2

        error_list = lag_mis["index"][mask].to_list()
        error_list.sort()
        return {"Missing": error_list}


def test_validate():
    import pandas as pd

    fake_mis = pd.DataFrame(
        [
            {
                "CHILD": "111",
                "MIS_START": "01/06/2020",
                "MIS_END": "05/06/2020",
            },  # 0 Fails, previous end is null
            {"CHILD": "111", "MIS_START": "04/06/2020", "MIS_END": pd.NA},  # 1
            {"CHILD": "222", "MIS_START": "03/06/2020", "MIS_END": "04/06/2020"},  # 2
            {"CHILD": "222", "MIS_START": "04/06/2020", "MIS_END": pd.NA},  # 3
            {
                "CHILD": "222",
                "MIS_START": "07/06/2020",
                "MIS_END": "09/06/2020",
            },  # 4 Fails, previous end is null
            {"CHILD": "333", "MIS_START": "02/06/2020", "MIS_END": "04/06/2020"},  # 5
            {
                "CHILD": "333",
                "MIS_START": "03/06/2020",
                "MIS_END": "09/06/2020",
            },  # 6 Fails, overlaps previous
            {"CHILD": "444", "MIS_START": "06/06/2020", "MIS_END": "08/06/2020"},  # 7
            {"CHILD": "444", "MIS_START": "10/06/2020", "MIS_END": "10/06/2020"},  # 8
            {
                "CHILD": "444",
                "MIS_START": "08/06/2020",
                "MIS_END": "10/08/2020",
            },  # 9 Fails, overlaps previous
            {"CHILD": "555", "MIS_START": pd.NA, "MIS_END": "05/06/2020"},  # 10
            {"CHILD": "555", "MIS_START": pd.NA, "MIS_END": "05/06/2020"},  # 11
        ]
    )

    fake_dfs = {"Missing": fake_mis}

    result = validate(fake_dfs)

    assert result == {"Missing": [0, 4, 6, 9]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="442",
    message="Unique Pupil Number (UPN) field is not completed.",
    affected_fields=["UPN"],
    tables=["Episodes", "Header"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("Header" not in dfs):
        return {}
    else:
        episodes = dfs["Episodes"]
        header = dfs["Header"]

        episodes.reset_index(inplace=True)
        header.reset_index(inplace=True)

        code_list = ["V3", "V4"]

        # merge to get all children for which episodes have been recorded.
        merged = episodes.merge(
            header, on=["CHILD"], how="inner", suffixes=["_eps", "_er"]
        )
        # Where any episode present, with an <LS> not = 'V3' or 'V4' then <UPN> must be provided
        mask = (~merged["LS"].isin(code_list)) & merged["UPN"].isna()
        header_error_locs = merged.loc[mask, "index_er"]

        return {
            # Select unique values since many episodes are joined to one header
            # and multiple errors will be raised for the same index.
            "Header": header_error_locs.dropna()
            .unique()
            .tolist()
        }


def test_validate():
    import pandas as pd

    fake_data_episodes = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108", "999"],
            "LS": [pd.NA, "L4", pd.NA, "L4", "L1", "V4", "V3", "XO"],
        }
    )
    fake_data_header = pd.DataFrame(
        {
            "CHILD": ["101", "102", "103", "104", "105", "106", "108"],
            "UPN": [pd.NA, "H801200001001", "UN1", "UN2", pd.NA, "UN3", pd.NA],
        }
    )
    fake_dfs = {"Episodes": fake_data_episodes, "Header": fake_data_header}

    result = validate(fake_dfs)

    assert result == {"Header": [0, 4]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="197a",
    message="Reason for no Strengths and Difficulties (SDQ) score is not required if Strengths and Difficulties Questionnaire score is filled in.",
    affected_fields=["SDQ_SCORE", "SDQ_REASON"],
    tables=["OC2"],
)
def validate(dfs):
    if "OC2" not in dfs:
        return {}
    else:
        oc2 = dfs["OC2"]

        sdq_filled_in = oc2["SDQ_SCORE"].notna()
        reason_filled_in = oc2["SDQ_REASON"].notna()

        error_mask = sdq_filled_in & reason_filled_in
        validation_error_locations = oc2.index[error_mask]

        return {"OC2": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "SDQ_SCORE": [0, pd.NA, 10, "1", pd.NA],
            "SDQ_REASON": ["XXX", pd.NA, pd.NA, "XXX", "XXX"],
        }
    )

    fake_dfs = {"OC2": fake_data}

    result = validate(fake_dfs)

    assert result == {"OC2": [0, 3]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="178",
    message="Placement provider code is not a valid code.",
    affected_fields=["PLACE_PROVIDER"],
    tables=["Episodes"],
)
def validate(dfs):
    if "Episodes" not in dfs:
        return {}

    episodes = dfs["Episodes"]

    code_list_placement_provider = ["PR0", "PR1", "PR2", "PR3", "PR4", "PR5"]
    code_list_placement_with_no_provider = ["T0", "T1", "T2", "T3", "T4", "Z1"]

    place_provider_needed_and_correct = episodes["PLACE_PROVIDER"].isin(
        code_list_placement_provider
    ) & ~episodes["PLACE"].isin(code_list_placement_with_no_provider)

    place_provider_not_provided = episodes["PLACE_PROVIDER"].isna()

    place_provider_not_needed = episodes["PLACE_PROVIDER"].isna() & episodes[
        "PLACE"
    ].isin(code_list_placement_with_no_provider)

    mask = (
        place_provider_needed_and_correct
        | place_provider_not_provided
        | place_provider_not_needed
    )

    validation_error_mask = ~mask
    validation_error_locations = episodes.index[validation_error_mask]

    return {"Episodes": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "PLACE_PROVIDER": ["PR0", "PR1", "", pd.NA, "", pd.NA, "pr4"],
            "PLACE": ["U1", "T0", "U2", "Z1", "T1", pd.NA, "U1"],
        }
    )

    fake_dfs = {"Episodes": fake_data}

    result = validate(fake_dfs)

    assert result == {"Episodes": [1, 2, 4, 6]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="221",
    message="The Ofsted Unique reference number (URN) provided for the child's placement does not match the placement postcode provided.",
    affected_fields=["PL_POST"],
    tables=["Episodes", "Provider Info"],
)
def validate(dfs):
    if ("Episodes" not in dfs) or ("provider_info" not in dfs["metadata"]):
        return {}
    else:
        episodes = dfs["Episodes"]
        provider_info = dfs["metadata"]["provider_info"]
        ls_list = ["V3", "V4"]
        place_list = ["H5", "K1", "K2", "R3", "S1"]
        xxx = "X" * 7
        # merge
        episodes["index_eps"] = episodes.index
        episodes = episodes[
            episodes["URN"].notna()
            & (episodes["URN"] != xxx)
            & (~episodes["LS"].isin(ls_list))
            & episodes["PLACE"].isin(place_list)
            & episodes["PL_POST"].notna()
        ]
        merged = episodes.merge(provider_info, on="URN", how="left")
        # If <URN> provided and <URN> not = 'XXXXXX', and <LS> not = 'V3', 'V4' and where <PL> = 'H5', 'K1', 'K2', 'R3' or 'S1' and <PL_POST> provided, <PL_POST> should = URN Lookup <Provider Postcode>
        mask = merged["PL_POST"].str.replace(" ", "") != merged["POSTCODE"]

        eps_error_locations = merged.loc[mask, "index_eps"]
        return {"Episodes": eps_error_locations.unique().tolist()}


def test_validate():
    import pandas as pd

    fake_data_eps = pd.DataFrame(
        [
            {
                "CHILD": "1111",
                "LS": "V3",
                "PLACE": "R3",
                "PL_POST": "A11 5KE",
                "URN": 1,
            },  # 0 ignore: LS is V3
            {
                "CHILD": "1111",
                "LS": "V2",
                "PLACE": "R3",
                "PL_POST": "PR5",
                "URN": pd.NA,
            },  # 1 ignore: URN value
            {
                "CHILD": "1111",
                "LS": "V2",
                "PLACE": "K1",
                "PL_POST": "S25 1WO",
                "URN": 3,
            },  # 2 fail
            {
                "CHILD": "2222",
                "LS": "V2",
                "PLACE": "K2",
                "PL_POST": "PR3",
                "URN": "XXXXXXX",
            },  # 3 ignore: URN value
            {
                "CHILD": "3333",
                "LS": "V2",
                "PLACE": "R3",
                "PL_POST": "S25 1WO",
                "URN": 2,
            },  # 4 pass
            {
                "CHILD": "3333",
                "LS": "V2",
                "PLACE": "xx",
                "PL_POST": "S25 1WO",
                "URN": 2,
            },  # 5 ignore: PLACE value
            {
                "CHILD": "4444",
                "LS": "V2",
                "PLACE": "S1",
                "PL_POST": "N9 5PY",
                "URN": 1,
            },  # 6 fail
            {
                "CHILD": "5555",
                "LS": "V2",
                "PLACE": "S1",
                "PL_POST": "N9 5PY",
                "URN": 4,
            },  # 7 fail
            {
                "CHILD": "5555",
                "LS": "V2",
                "PLACE": "R3",
                "PL_POST": pd.NA,
                "URN": 1,
            },  # 8 ignore: PL_POST value
        ]
    )
    fake_provider_info = pd.DataFrame(
        [
            {
                "URN": 1,
                "POSTCODE": "A115KE",
            },  # 0
            {
                "URN": 2,
                "POSTCODE": "S251WO",
            },  # 1
            {
                "URN": 3,
                "POSTCODE": "V29XX",
            },  # 2
            {
                "URN": 4,
                "POSTCODE": pd.NA,
            },  # 3 should NaNs be ignored?
        ]
    )
    metadata = {"provider_info": fake_provider_info}

    fake_dfs = {"Episodes": fake_data_eps, "metadata": metadata}

    result = validate(fake_dfs)

    assert result == {"Episodes": [2, 6, 7]}



import pandas as pd

from lac_validator.rule_engine import rule_definition


@rule_definition(
    code="621",
    message="Mother’s field has been completed but date of birth shows that the mother is younger than her child.",
    affected_fields=["DOB", "MC_DOB"],
    tables=["Header"],
)
def validate(dfs):
    if "Header" not in dfs:
        return {}
    else:
        header = dfs["Header"]

        header["MC_DOB"] = pd.to_datetime(
            header["MC_DOB"], format="%d/%m/%Y", errors="coerce"
        )
        header["DOB"] = pd.to_datetime(
            header["DOB"], format="%d/%m/%Y", errors="coerce"
        )
        mask = (header["MC_DOB"] > header["DOB"]) | header["MC_DOB"].isna()

        validation_error_mask = ~mask
        validation_error_locations = header.index[validation_error_mask]

        return {"Header": validation_error_locations.tolist()}


def test_validate():
    import pandas as pd

    fake_data = pd.DataFrame(
        {
            "DOB": [
                "01/12/2021",
                "19/02/2016",
                "31/01/2019",
                "31/01/2019",
                "31/01/2019",
            ],
            "MC_DOB": ["01/01/2021", "19/12/2016", "31/01/2019", pd.NA, "01/02/2019"],
        }
    )

    fake_dfs = {"Header": fake_data}

    result = validate(fake_dfs)

    assert result == {"Header": [0, 2]}
